/* $EPIC:$ */
/*
 * window.h: header file for window.c 
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993, 1997 Jeremy Nelson.
 * Copyright  1994 Jake Khuon.
 * Copyright  1995, 2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#ifndef __window_h__
#define __window_h__

typedef int	Winref;

/*
 * WindowStack: The structure for the POP, PUSH, and STACK functions. A
 * simple linked list with window refnums as the data 
 */
typedef	struct	window_stack_stru
{
	Winref				refnum;
	struct	window_stack_stru	*next;
}	WindowStack;


#define NEED_WINDOW_INTERNALS
#ifdef NEED_WINDOW_INTERNALS

/* To get the definition of Lastlog */
#include "lastlog.h"

/* To get the definition of Status */
#include "status.h"

/* 
 * Screen and Window are mutually referencial.
 * That means we cant include "screen.h", so we kind of fake it.
 */
struct	ScreenStru;

/* Flags for "misc" */
#define WINDOW_NOTIFY		1 << 0
#define WINDOW_NOTIFIED		1 << 1

/* Should be a way to make static to window.c */
typedef	struct	DisplayStru
{
	char			*line;
	struct	DisplayStru	*prev;
	struct	DisplayStru	*next;
}	Display;

typedef struct	WNickListStru
{
struct WNickListStru	*next;
	char		*nick;
} WNickList;

typedef	struct	WindowStru
{
	Winref	refnum;			/* Unique refnum for window */
	char	*name;			/* Logical name for window */
	int	server;			/* Server that win is connected to */
	int	last_server;		/* Last server connected to */
	int	top;			/* SCREEN line for top of window */
	int	bottom;			/* SCREEN line for bottom of window */
	int	cursor;			/* WINDOW line where the cursor is */
	int	line_cnt;		/* Number of lines ON DISPLAY */
	int	absolute_size;		/* True if window doesnt rebalance */
	int	scroll;			/* True if the window scrolls */
	int	scratch_line;		/* True if this is a scratch window */
	int	noscroll;		/* True if window doesn't scroll */
	int	old_size;		/* 
					 * Usu. same as display_size except
					 * right after a screen resize, and
					 * is used as a flag for resize_display
					 */
	int	visible;		/* True if the window is on a SCREEN */
	int	update;			/* True if window display is dirty */
	int	repaint_start;		/* Where to start repainting */
	int	repaint_end;		/* Where to stop repainting */
	unsigned miscflags;		/* Miscellaneous flags. */
	int	beep_always;		/* True if a beep to win always beeps */
	int	notify_level;		/* the notify level.. */
	int	window_level;		/* Lastlog level for the window */
	int	skip;			/* Whether window should be skipped */
	int	columns;		/* How wide we are when hidden */

	/* Input and Status stuff */
	char	*prompt;		/* Current EXEC prompt for window */
	Status	status;			/* Current status line info */

	/* Display stuff */
	/*
	 * The "scrollback" buffer is a linked list of lines that have
	 * appeared, are appearing, or will appear on the visible window.
	 * The "top" of the scrollback buffer usually floats forward in
	 * the line as items are added.  The visible part of the screen is
	 * always somewhere in the buffer (usually at the bottom), unless
	 * the user is in "scrollback mode" or "hold mode".  When the user
	 * is in either of these modes, then the top of the scrollback buffer
	 * is frozen (so as not to disrupt the current contents of the screen)
	 * and the bottom of the buffer floats as is neccesary.  When the user
	 * ends these modes, the buffer is reset to the "scrollback_point" and
	 * then is floated to its original dimensions from there.  The lastlog
	 * has nothing to do with scrollback any longer.
	 *
	 * The "display_ip" is always a blank line where the NEXT displayable
	 * line will go.  When the user does a /clear, the screen is scrolled
	 * up until display_ip is the top_of_display.  "display_size" is the
	 * number of rows that can appear on the screen at a given time.
	 */
	Display *top_of_scrollback,	/* Start of the scrollback buffer */
		*top_of_display,	/* Where the viewport starts */
		*ceiling_of_display,	/* The farthest back top_of_display */
		*display_ip,		/* Where next line goes in rite() */
		*scrollback_point;	/* Where t_o_d was at start of sb */
	int	display_buffer_size;	/* How big the scrollback buffer is */
	int	display_buffer_max;	/* How big its supposed to be */
	int	display_size;		/* How big the window is - status */

	int	lines_scrolled_back;	/* Where window is relatively */

	int	hold_mode;		/* True if we want to hold stuff */
	int	holding_something;	/* True if we ARE holding something */
	int	held_displayed;		/* lines displayed since last hold */
	int	lines_displayed;	/* Lines held since last unhold */
	int	lines_held;		/* Lines currently being held */
	int	last_lines_held;	/* Last time we updated "lines held" */
	int	distance_from_display;	/* How far t_o_d is from d_ip */

	/* Channel stuff */
	char	*query_nick;		/* Current default target for win */
	WNickList *nicks;		/* List of nick-queries for this win */


	/* /LASTLOG stuff */
	Lastlog	*lastlog_head;		/* pointer to top of lastlog list */
	Lastlog	*lastlog_tail;		/* pointer to bottom of lastlog list */
	int	lastlog_level;		/* The LASTLOG_LEVEL, determines what
					 * messages go to lastlog */
	int	lastlog_size;		/* number of messages in lastlog. */
	int	lastlog_max;		/* Max number of messages in lastlog */


	/* /WINDOW LOG stuff */
	/* XXX 'log_refnum' should be ``Logref'', but its too hard for now */
	int	log_refnum;		/* Our logging handle */
	int	logging;		/* True if logging is ON */
	char	*logfile;		/* window's logfile name */

	/* List stuff */
struct	ScreenStru	*screen;	/* The screen we belong to */
struct	WindowStru	*next;		/* Window below us on screen */
struct	WindowStru	*prev;		/* Window above us on screen */

	/* XXX - Help stuff */
struct	HelpStru	*helper;	/* Info about current /help */

	int		deceased;	/* Set when the window is killed */

	void		(*add) (struct WindowStru *, const u_char *);
	void		(*lastlog) (struct WindowStru *, const u_char *);
	void		(*display) (struct WindowStru *, const u_char *);

}	Window;


extern	Window *	current_window;
extern	Window *	invisible_list;

	/* XXX Some of these should be made static */
	Window *	new_window 			(struct ScreenStru *);
	void		delete_window			(Window *);
	int		traverse_all_windows		(Window **);
	void		add_to_invisible_list		(Window *);
	Window *	add_to_window_list	(struct ScreenStru *, Window *);
	void		remove_window_from_screen	(Window *);
	void		recalculate_window_positions	(struct ScreenStru *);
	void		rebalance_windows		(struct ScreenStru *);
	void		recalculate_windows		(struct ScreenStru *);
	Window *	get_window_by_desc		(Char *);
	Window *	get_window			(Winref);
	Window *	get_window_by_name		(Char *);
	void		hide_window			(Window *);
	void		update_window_status		(Window *, int);
	int		add_to_display_list	(Window *, const u_char *);
	int		number_of_windows_on_screen	(Window *);
	char *		get_nicklist_by_window		(Window *); /* XXX */
	Window *	window_channel			(Window *, char **);
	Window *	window_query			(Window *, char **);
	Window *	window_scroll			(Window *, char **);
	Window *	window_server			(Window *, char **);
	void		hold_mode			(Window *, int , int);
	void		flush_everything_being_held	(Window *);
	int		unhold_a_window			(Window *);
	void		recalculate_window_cursor	(Window *);
	void		make_window_current		(Window *);

#endif /* NEED_WINDOW_DEFINITION */

	void		delete_all_windows		(void);
	int		traverse_all_windows_logical	(Winref *);
	void		redraw_all_windows		(void);
	void		update_all_windows		(void);
	void		status_needs_updating		(Winref);
	void		swap_last_window		(char, char *);
	void		next_window			(char, char *);
	void		swap_next_window		(char, char *);
	void		previous_window			(char, char *);
	void		swap_previous_window		(char, char *);
	int		is_window_visible		(char *);
	Winref		get_current_winref		(void);
	Winref		get_winref_by_desc		(Char *);
	Char *		get_window_name			(Winref);
	char *		get_status_by_winref		(Winref, int);
	void		update_all_status		(void);
	int		status_update			(int);
	void		set_prompt_by_winref		(Winref, char *);
	char *		get_prompt_by_winref		(Winref);
	Char *		get_target_by_winref		(Winref);
	Char *		query_nick			(void);
	Char *		get_window_channel		(Winref);
	int		get_window_server		(Winref);
	int		get_window_oldserver		(Winref);
	void		change_window_server		(int, int);
	int		windows_connected_to_server	(int);
	void		window_check_servers		(void);
	void		set_level_by_winref		(Winref, int);
	void		clear_all_windows		(int, int, int);
	void		clear_window_by_winref		(Winref);
	void		unclear_all_windows		(int, int, int);
	void		unclear_window_by_winref	(Winref);
	void		set_scroll_lines		(int);
	void		set_continued_line		(char *);
	void		set_scrollback_size		(int);
	int		get_geom_by_winref		(Winref, int *, int *);
	int		get_display_size_by_winref	(Winref);
	char *		get_nicklist_by_winref		(Winref);
	Char *		get_query_by_winref		(Winref);
	int		get_level_by_winref		(Winref);
	int		get_scratch_window_value	(Winref);
	void		set_window_scratch_line		(Winref, int);
	BUILT_IN_COMMAND(windowcmd);
	void		scrollback_backwards		(char, char *);
	void		scrollback_forwards		(char, char *);
	void		scrollback_end			(char, char *);
	void		scrollback_start		(char, char *);
	int		unhold_windows			(void);
	void		unstop_all_windows		(char, char *);
	void		reset_line_cnt			(int);
	void		toggle_stop_screen		(char, char *);
	void		make_window_current_by_desc	(Char *);
	void		make_window_current_by_winref	(int);

extern	int		in_window_command;
extern	unsigned 	window_display;

#endif /* __window_h__ */
