/* $EPIC:$ */
/*
 * log.c: All of the high level logging stuff for ircII.  Uses the new
 * 	functions provided by nutty.c
 *
 * Copyright  2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

/* XXXX This file just works for now, it needs a major league cleanup. */
/* 
 * XXXX Yes, i know i do things intentionally the slow way.  I'll fix 
 * that in the future once i know this all works.  Be patient please.
 */

#include "irc.h"
#include "window.h"
#include "server.h"
#include "ircaux.h"
#include "vars.h"
#include "output.h"
#include "nutty.h"
#include "log.h"
#include "alias.h"
#include <sys/stat.h>

/**************************** NON GLOBAL LOGS ******************************/
/*
 * This is the general purpose filtered logging support, intended to support
 * the /window log, /window logfile, and /log commands.  Other places should
 * also find this relatively easy to use.  This is layered upon the very
 * low level nutty interface, rather than on the higher level nutty interface
 * because 1) it was easier this way 2) I wanted to get this working, and
 * the high level nutty interface wasn't finished when i ran out of time.
 * Maybe some day in the future i'll change this implementation, but when
 * that day comes, you shouldn't notice a thing.  Just use the extern funcs
 * and all will be well.
 */
typedef struct logs {
	Logref		refnum;		/* Unique ID for this log */
	int		active;		/* Someone has a handle on this log */
	output_info 	info;		/* Information about what to log */
	char *		rewrite;	/* Code used to rewrite the output */
	int		mangler;	/* Flags used to mangle the output */
	char *		filename;	/* Filename to use for the log */
	FILE *		fp;		/* File to write to, NULL if log off */
} Log;

static	Log *	log_info;
static	int	log_info_size = 0;
static	int	num_logs = 0;

static 	Log *	get_log_by_handle 	(Logref);
static 	void	clean_log 		(Logref, int);
static	void	resize_logs 		(void);
static	void	general_logger 		(const output_info *, const u_char *);
static	Char *	log_time		(void);


/* Private logging helper functions */
static Log *	get_log_by_handle (Logref l)
{
	int	i;

	for (i = 0; i < num_logs; i++)
	{
		if (log_info[i].refnum == l)
		{
			if (log_info[i].active)
				return &log_info[i];
			else
				return NULL;
		}
	}

	return NULL;
}

static void	clean_log (Logref i, int newclean)
{
	log_info[i].refnum = i;
	log_info[i].active = 0;
	log_info[i].info.winref = -1;
	log_info[i].info.servref = -1;
	log_info[i].info.who_level = 0;
	log_info[i].mangler = 0;

	if (newclean)
	{
		log_info[i].info.who_from = NULL;
		log_info[i].rewrite = NULL;
		log_info[i].filename = NULL;
		log_info[i].fp = NULL;
	}
	else
	{
		new_free(&log_info[i].info.who_from);
		new_free(&log_info[i].rewrite);
		new_free(&log_info[i].filename);
		if (log_info[i].fp)
			fclose(log_info[i].fp);
		log_info[i].fp = NULL;
	}
}

static void	resize_logs (void)
{
	int	old = log_info_size;
	int	i;

	if (log_info_size == 0)
		register_output_callback(general_logger);

	log_info_size += 64;
	RESIZE(log_info, Log, log_info_size);
	for (i = old; i  < log_info_size; i++)
		clean_log(i, 1);
}

/* Callback for all types of logging */
static void	general_logger (const output_info *info, const u_char *output)
{
	int	j;

	for (j = 0; j < num_logs; j++)
	{
		output_info *	i;
		u_char *	my_output = NULL;
		int		must_free = 0;

		if (!log_info[j].active)
			continue;
		if (!log_info[j].fp)
			continue;

		i = &log_info[j].info;
		if (i->winref != -1 && i->winref != info->winref)
			continue;
		if (i->servref != -1 && i->servref != info->servref)
			continue;
		if (info->who_from == NULL && i->who_from != NULL)
			continue;
		if (info->who_from != NULL && i->who_from == NULL)
			continue;
		if (info->who_from && i->who_from &&
				my_stricmp(i->who_from, info->who_from))
			continue;
		if (i->who_level && ((i->who_level & info->who_level) == 0))
			continue;

		my_output = LOCAL_COPY(output);

		/* Permit the user to rewrite the line... */
		if (log_info[j].rewrite)
		{
			char *	args;
			char *	stuff;
			char *	result;

			stuff = LOCAL_COPY(log_info[j].rewrite);
			args = LOCAL_COPY(my_output);
			result = parse_line_with_return("genlog", stuff, 
							args, 0, 0);
			if (must_free)
				new_free(&my_output);
			my_output = result;
			must_free = 1;
		}

		/* Mangle the line... */
		if (log_info[j].mangler)
		{
			size_t	count = strlen(my_output) * 3 + 1;
			char *	my_output2;

			my_output2 = (u_char *)alloca(count + 1);
			strcpy(my_output2, my_output);
			mangle_line(my_output2, log_info[j].mangler, count);

			if (must_free)
				malloc_strcpy((char **)&my_output, my_output2);
			else
				my_output = my_output2;
		}
		else
			my_output = LOCAL_COPY(output);

		/* Then output it. */
		fprintf(log_info[j].fp, "%s\n", my_output);
		fflush(log_info[j].fp);

		if (must_free)
			new_free(&my_output);
	}
}

static	const char *	log_time (void)
{
	time_t		t;
	static char 	my_buffer[256];
	struct tm *	ugh;

	time(&t);
	ugh = localtime(&t);
	strftime(my_buffer, 255, "%a %b %d %H:%M:%S %Y", ugh);
	return my_buffer;
}



/* Public logging API */
Logref	get_new_log_handle (void)
{
	int	i;

	for (i = 0; ; resize_logs())
	{
		for (; i < log_info_size; i++)
		{
			if (log_info[i].active == 0)
			{
				clean_log(i, 0);
				log_info[i].active = 1;
				if (i >= num_logs)
					num_logs = i + 1;
				return i;
			}
		}
	}
}

int	change_log_info (Logref logref, const output_info *info)
{
	Log *	log;

	if (!(log = get_log_by_handle(logref)))
		return -1;

	log->info.winref = info->winref;
	log->info.servref = info->servref;
	malloc_strcpy((char **)&log->info.who_from, info->who_from);
	log->info.who_level = info->who_level;
	return 0;
}

int	change_log_rewrite (Logref logref, const char *rewrite)
{
	Log *	log;

	if (!(log = get_log_by_handle(logref)))
		return -1;
	malloc_strcpy(&log->rewrite, rewrite);
	return 0;
}

int	change_log_mangler (Logref logref, int mangler)
{
	Log *	log;

	if (!(log = get_log_by_handle(logref)))
		return -1;
	log->mangler = mangler;
	return 0;
}

int	change_log_filename (Logref logref, const char *filename)
{
	Log *	log;
	char *	str;
	char *	expanded;
	char *	newfilename;
	int	cycle;

	if (!(log = get_log_by_handle(logref)))
		return -1;

	str = LOCAL_COPY(filename);
	if (!(expanded = expand_twiddle(str)))
	{
		say("LOG: No such user");
		return -1;
	}
	newfilename = LOCAL_COPY(expanded);
	new_free(&expanded);

	if (log->filename && !strcmp(log->filename, newfilename))
		return 0;		/* No change in filename */

	cycle = (log->fp) ? 1 : 0;

	if (cycle)
		turn_log_off(logref);
	malloc_strcpy(&log->filename, newfilename);
	if (cycle)
		turn_log_on(logref);
	return 0;
}

int	turn_log_on (Logref logref)
{
	Log *	log;

	if (!(log = get_log_by_handle(logref)))
		return -1;

	if (log->fp)
		return -1;	/* Already on */

	if (!log->filename)
	{
		say("Log %d needs a filename before it can be turned on",
			logref);
		return -1;
	}

	if (!(log->fp = fopen(log->filename, "a")))
	{
		say("Couldn't open logfile %s: %s", 
			log->filename, strerror(errno));
		return -1;
	}

	say("Starting logfile %s", log_time());
	fchmod(fileno(log->fp), S_IREAD | S_IWRITE);
	fprintf(log->fp, "IRC log started %s\n", log_time());
	fflush(log->fp);
	return 0;
}

int	turn_log_off (Logref logref)
{
	Log *	log;

	if (!(log = get_log_by_handle(logref)))
		return -1;

	if (!log->fp)
		return -1;

	fprintf(log->fp, "IRC log ended %s\n", log_time());
	fflush(log->fp);
	fclose(log->fp);
	log->fp = NULL;
	return 0;
}

int	release_log (Logref logref)
{
	Log *	log;

	if (!(log = get_log_by_handle(logref)))
		return -1;

	if (!log->active)
		return -1;

	if (log->fp)
		turn_log_off(logref);

	clean_log(logref, 0);
	return 0;
}

void	cancel_all_logs (void)
{
	int	i;

	for (i = 0; i < num_logs; i++)
	{
		if (log_info[i].active)
			release_log(log_info[i].refnum);
	}
}

/************************** THE /LOG COMMAND *********************************/
static void	log_info_one (Log *l)
{
	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return;
	}

	say("Log refnum [%d], is active? [%s]", 
		l->refnum, l->active ? "YES" : "NO");
	say("Info.winref [%d]", l->info.winref);
	say("Info.servref [%d]", l->info.servref);
	say("Info.who_from [%s]", l->info.who_from ? 
				   l->info.who_from : "NULL");
	say("Info.who_level [%d]", l->info.who_level);
	say("rewriter: [%s]", l->rewrite ? l->rewrite : "NULL");
	say("log open? [%s]", l->fp ? "YES" : "NO");

	return;
}

static Log *	log_filename (Log *l, char **args)
{
	char *	arg;

	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	if ((arg = new_next_arg(*args, args)))
	{
		change_log_filename(l->refnum, arg);
		return l;
	}

	say("LOG REFNUM requires an argument.");
	return NULL;
}

static Log *	log_info_func (Log *l, char **args)
{
	int	i;

	if (l)
	{
		log_info_one(l);
		return l;
	}

	say("Currently active logs:");
	for (i = 0; i < num_logs; i++)
		log_info_one(&log_info[i]);
	say("End of Log information.");

	return l;
}

static Log *	log_list (Log *k, char **args)
{
	int	i;

	say("%2.2s %2.2s %2.2s %-10.10s %-10.10s %s",
		"R",  "W",  "S",  "T",    "F",  "L" );

	for (i = 0; i < num_logs; i++)
	{
		Log *	l = &log_info[i];

		if (!l->active)
			continue;

		say("%2d %2d %2d %-10.10s %-10.10s %s",
			l->refnum, l->info.winref, l->info.servref,
			l->info.who_from ? l->info.who_from : "NONE",
			l->filename,
			bits_to_lastlog_level(l->info.who_level));
	}

	return k;
}

static Log *	log_new (Log *l, char **args)
{
	return get_log_by_handle(get_new_log_handle());
}

static Log *	log_refnum (Log *l, char **args)
{
	char *	arg;
	Log *	retval;
	Logref	ref;

	if ((arg = next_arg(*args, args)))
	{
		if (is_number(arg))
		{
			ref = (Logref)my_atol(arg);
			if ((retval = get_log_by_handle(ref)))
				return retval;
		}

		say("LOG REFNUM \"%s\" is not a valid active log refnum", arg);
		return NULL;
	}

	say("LOG REFNUM requires an argument.");
	return NULL;
}

static Log *	log_winref (Log *l, char **args)
{
	char *	arg;
	Winref	ref;

	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	if ((arg = new_next_arg(*args, args)))
	{
		if ((ref = get_winref_by_desc(arg)) != -1)
		{
			l->info.winref = ref;
			return l;
		}

		say("\"%s\" is not a valid window description", arg);
		return NULL;
	}

	say("LOG WINDOW requires an argument.");
	return NULL;
}

static Log *	log_servref (Log *l, char **args)
{
	char *	arg;
	int	ref;

	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	if ((arg = new_next_arg(*args, args)))
	{
		if ((ref = get_servref_by_desc(arg)) != -1)
		{
			l->info.servref = ref;
			return l;
		}

		say("\"%s\" is not a valid window description", arg);
		return NULL;
	}

	say("LOG WINDOW requires an argument.");
	return NULL;
}

static Log *	log_target (Log *l, char **args)
{
	char *	arg;

	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	if ((arg = new_next_arg(*args, args)))
	{
		malloc_strcpy((char **)&l->info.who_from, arg);
		return l;
	}

	say("LOG TARGET requires an argument.");
	return NULL;
}

static Log *	log_level (Log *l, char **args)
{
	char *	arg;
	int	level;

	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	if ((arg = new_next_arg(*args, args)))
	{
		level = parse_lastlog_level(arg);
		l->info.who_level = level;
		return l;
	}

	say("LOG LEVEL requires an argument.");
	return NULL;
}

static Log *	log_rewrite (Log *l, char **args)
{
	char *	arg;
	char *	after = NULL;

	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	if (**args == '{')
	{
		if ((after = MatchingBracket(*args + 1, '{', '}')))
		{
			arg = *args;
			*after++ = 0;
			*args = after;
		}
	}
	else if (!(arg = new_next_arg(*args, args)))
	{
		say("LOG LEVEL requires an argument.");
		return NULL;
	}

	change_log_rewrite(l->refnum, arg);
	return l;
}

static Log *	log_mangler (Log *l, char **args)
{
	char *	nv = NULL;
	char *	arg;
	int	mangle;

	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	mangle = l->mangler;
	if ((arg = new_next_arg(*args, args)))
	{
		mangle = parse_mangle(arg, mangle, &nv);
		change_log_mangler(l->refnum, mangle);

		new_free(&nv);
		return l;
	}

	say("LOG MANGLE requires an argument.");
	return NULL;
}

static Log *	log_on (Log *l, char **args)
{
	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	turn_log_on(l->refnum);
	return l;
}

static Log *	log_off (Log *l, char **args)
{
	if (!l)
	{
		say("LOG: Must specify a log refnum first.");
		return NULL;
	}

	turn_log_off(l->refnum);
	return l;
}

typedef Log *(*log_func) (Log *, char **args);

typedef struct log_ops_T {
	char *		command;
	log_func	func;
} log_ops;

static log_ops options [] = {
	{ "FILENAME",		log_filename	},
	{ "INFO",		log_info_func	},
	{ "LEVEL",		log_level	},
	{ "LIST",		log_list	},
	{ "MANGLE",		log_mangler	},
	{ "NEW",		log_new 	},
	{ "OFF",		log_off		},
	{ "ON",			log_on		},
	{ "REFNUM",		log_refnum	},
	{ "REWRITE",		log_rewrite	},
	{ "SERVER",		log_servref	},
	{ "TARGET",		log_target	},
	{ "WINDOW",		log_winref	},
	{ NULL,			NULL		}
};

BUILT_IN_COMMAND(logcmd)
{
	char *	arg;
	int	nargs = 0;
	Log *	log = NULL;

	while ((arg = next_arg(args, &args)))
	{
		int	i;
		int	len = strlen(arg);

		if (*arg == '-' || *arg == '/')
			arg++, len--;

		for (i = 0; options[i].func; i++)
		{
			if (!my_strnicmp(arg, options[i].command, len))
			{
				nargs++;
				if (!(log = options[i].func(log, &args)))
					args = NULL;
				break;
			}
		}
	}

	if (!nargs)
		log_list(NULL, NULL);
}

/************************* GLOBAL LOGFILE SUPPORT ****************************/
/*
 * This is the global logfile support, ie, /SET LOG ON/OFF, /SET LOGFILE <x>.
 * This is implemented on top of the low level nutty interface as a proof of
 * concept of how to use the low level nutty interface.  Basically, when you
 * /SET LOG ON, we open up the global log file, and register ``global_log_it''
 * to recieve copies of all output for the client.  Then whenever global_log_it
 * is called, we just dump that line of output to the logfile.  Of course, 
 * we honor /SET MANGLE_LOGFILE.
 *
 * When you /SET LOG OFF, we deregister global_log_it(), and thus no more
 * logging takes place.  But even if we did slip up, global_log_it() will
 * refuse to log if the logfile doesn't exist, just as a safety precaution.
 *
 * If you /SET LOGFILE when /SET LOG ON, then we have to do an implied
 * /SET LOG OFF before we change the log file, then /SET LOG ON again.
 * If you /SET LOGFILE to the same nickname when /SET LOG ON, then we have
 * to honor that as a no-op.  What a pain.  Please don't do this.
 * If you /SET LOGFILE when /SET LOG OFF, then we don't need to go through
 * any of that hassle.
 */
static	Logref		global_logref = -1;
static	int		global_logging = 0;
static	int		global_log_mangler = 0;

/* This function is called when you do /SET LOGFILE */
void	set_logfile_callback (const char *fn)
{
	if (global_logref == -1)
		panic("/SET LOG should have been called during init");

	change_log_filename(global_logref, fn);
}

/* This function is called when you do /SET LOG */
void	set_log_callback (int newval)
{
	/* Initial bootup at client start */
	if (global_logref == -1)
		global_logref = get_new_log_handle();

	/* Doing a /SET LOG ON when LOG is OFF */
	else if (global_logging == 0 && newval == 1)
		turn_log_on(global_logref);

	/* Doing a /SET LOG OFF when LOG is ON */
	else if (global_logging == 1 && newval == 0)
		turn_log_off(global_logref);

	global_logging = newval;
}

/* This function is called when you do /SET MANGLE_LOGFILES */
void	set_mangle_logfiles (const char *value)
{
	char *	nv = NULL;

	global_log_mangler = parse_mangle(value, global_log_mangler, &nv);
	set_string_var(MANGLE_LOGFILES_VAR, nv);
	new_free(&nv);

	change_log_mangler(global_logref, global_log_mangler);
}


